/*
 *  Copyright (c) 2003 Helsinki Institute for Information Technology
 *
 *  Permission is hereby granted, free of charge, to any person obtaining a copy
 *  of this software and associated documentation files (the "Software"), to
 *  deal in the Software without restriction, including without limitation the
 *  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 *  sell copies of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *
 *  The above copyright notice and this permission notice shall be included in
 *  all copies or substantial portions of the Software.
 *
 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 *  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 *  IN THE SOFTWARE.
 */
 
package vbsearch;

import com.hp.hpl.jena.ontology.*;
import com.hp.hpl.jena.rdf.model.*;
import com.hp.hpl.jena.shared.PrefixMapping;
import java.io.*;
import java.net.URISyntaxException;
import java.util.*;
import java.util.regex.*;
import javax.xml.parsers.*;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.sax.SAXTransformerFactory;
import javax.xml.transform.sax.TransformerHandler;
import javax.xml.transform.stream.StreamResult;
import org.xml.sax.*;
import org.xml.sax.helpers.*;
import seco.io.*;
import seco.util.iterator.*;
import vbsearch.impl.jena.*;
import vbsearch.impl.mem.*;
import vbsearch.query.*;
import vbsearch.util.*;

/**
 *  Description of the Class
 *
 *@author     Samppa
 *@created    September 15, 2003
 */
public class VBSearch {
    private final Attributes EMPTY_ATTRIBUTES = new AttributesImpl();
    private final Attributes PARSE_TYPE_COLLECTION = new AttributesImpl();
    /**
     *  Description of the Field
     */
    private int anonID = 0;

    /**
     *  Description of the Field
     */
    private Map anonMap = new HashMap();
    /**
     *  Description of the Field
     */
    private VBSIndex index;

    private PrefixMapping prefixMapping = PrefixMapping.Factory.create();

    private SAXTransformerFactory tfactory = null;
    {
        prefixMapping.setNsPrefix("ogt", VBSConf.OGT_NS);
        prefixMapping.setNsPrefix("bm", VBSConf.BM_NS);
        prefixMapping.setNsPrefix("rdf", VBSConf.RDF_NS);
        prefixMapping.setNsPrefix("rdfs", VBSConf.RDFS_NS);

        ((AttributesImpl) PARSE_TYPE_COLLECTION).addAttribute(VBSConf.RDF_NS, "parseType", "rdf:parseType", "CDATA", "Collection");
    }


    /**
     *  Constructor for the VBSearch object
     */
    public VBSearch() {
        this(null);
    }


    /**
     *  Constructor for the VBSearch object
     *
     *@param  index  Description of the Parameter
     */
    public VBSearch(VBSIndex index) {
        tfactory = (SAXTransformerFactory) SAXTransformerFactory.newInstance();
        if (index != null) {
            this.index = index;
        } else {
            this.index = new VBSIndexImpl();
        }
    }


    /**
     *  Description of the Method
     *
     *@param  model                         Description of the Parameter
     *@param  document                      Description of the Parameter
     *@exception  UnknownResourceException  Description of the Exception
     *@exception  URISyntaxException        Description of the Exception
     */
    public void add(OntModel model, String document) throws UnknownResourceException, URISyntaxException {
        add(new JenaInputLoader(model, document));
        System.gc();
    }


    /**
     *  Description of the Method
     *
     *@param  inputLoader                   Description of the Parameter
     *@exception  UnknownResourceException  Description of the Exception
     *@exception  URISyntaxException        Description of the Exception
     */
    public void add(VBSInputLoader inputLoader) throws UnknownResourceException, URISyntaxException {
        PrefixMapping pm = inputLoader.getPrefixMapping();
        this.prefixMapping.setNsPrefixes(pm);
        this.index.setCategoryBase(inputLoader.getCategoryBase());
        for (Iterator roots = inputLoader.getRootCategories(); roots != null && roots.hasNext(); ) {
            add((Category) roots.next());
        }
        for (Iterator bookmarks = inputLoader.getBookmarks(); bookmarks != null && bookmarks.hasNext(); ) {
            add((Bookmark) bookmarks.next());
        }
    }


    /**
     *  Description of the Method
     *
     *@param  category                      Description of the Parameter
     *@exception  UnknownResourceException  Description of the Exception
     *@exception  URISyntaxException        Description of the Exception
     */
    public void add(Category category) throws UnknownResourceException, URISyntaxException {
        this.index.addCategory(category);
    }


    /**
     *  Description of the Method
     *
     *@param  bookmark                      Description of the Parameter
     *@exception  UnknownResourceException  Description of the Exception
     *@exception  URISyntaxException        Description of the Exception
     */
    public void add(Bookmark bookmark) throws UnknownResourceException, URISyntaxException {
        this.index.addBookmark(bookmark);
    }


    /**
     *  Adds a feature to the CategoryPath attribute of the VBSearch object
     *
     *@param  pathIter               The feature to be added to the CategoryPath
     *      attribute
     *@param  incCategoryProperties  The feature to be added to the CategoryPath
     *      attribute
     *@param  intersection           The feature to be added to the CategoryPath
     *      attribute
     *@param  incBookmarkHits        The feature to be added to the CategoryPath
     *      attribute
     *@param  contentHandler         The feature to be added to the CategoryPath
     *      attribute
     *@param  lang                   The feature to be added to the CategoryPath
     *      attribute
     *@param  incSubCategoryHits
     *@exception  SAXException       Description of the Exception
     */
    public void addCategoryPath(Iterator pathIter,
            Collection incCategoryProperties,
            Collection intersection,
            boolean incBookmarkHits,
            boolean incSubCategoryHits,
            ContentHandler contentHandler,
            String lang) throws SAXException {
        Category category = (Category) pathIter.next();
        startCategoryElement(category,
                incCategoryProperties,
        // incCategoryProperties
                null,
        // incBookmarkProperties
                intersection,
                incBookmarkHits,
                false,
        // incPathBookmarkHits
                false,
        // incDirectBookmarkHits
                false,
        // incPath
                FacetSelector.NONE,
        //incSubCategories
                FacetSelector.NONE,
        // filter
                false,
        // incBookmarks
                FacetSelector.NONE,
        // incTopics
                0,
        // maxSubCategories
                0,
        // maxBookmarks
                null,
        // maxSubCategoriesMap
                null,
        // maxBookmarksMap
                incSubCategoryHits,
                false,
        // incPathSubCategoryHits
                null,
        // skipSubCategoriesMap
                null,
        // skipBookmarksMap,
                false,
        // incPosition
                contentHandler,
                lang,
                null);
        // groupedBookmarks
        if (pathIter.hasNext()) {
            contentHandler.startElement(VBSConf.OGT_NS, "subCategories", this.prefixMapping.getNsURIPrefix(VBSConf.OGT_NS) + ":subCategories", PARSE_TYPE_COLLECTION);
            addCategoryPath(pathIter,
                    incCategoryProperties,
                    intersection,
                    incBookmarkHits,
                    incSubCategoryHits,
                    contentHandler,
                    lang);
            contentHandler.endElement(VBSConf.OGT_NS, "subCategories", this.prefixMapping.getNsURIPrefix(VBSConf.OGT_NS) + ":subCategories");
        }
        contentHandler.endElement(VBSConf.OGT_NS, "Category", this.prefixMapping.getNsURIPrefix(VBSConf.OGT_NS) + ":Category");
    }


    /**
     *  Adds a feature to the Properties attribute of the VBSearch object
     *
     *@param  resource          The feature to be added to the Properties
     *      attribute
     *@param  incProperties     The feature to be added to the Properties
     *      attribute
     *@param  contentHandler    The feature to be added to the Properties
     *      attribute
     *@param  lang              The feature to be added to the Properties
     *      attribute
     *@exception  SAXException  Description of the Exception
     */
    private void addProperties(VBSResource resource, Collection incProperties, ContentHandler contentHandler, String lang) throws SAXException {
        if (incProperties != null) {
            for (Iterator iter = incProperties.iterator(); iter.hasNext(); ) {
                String property = iter.next().toString();
                try {
                    addProperty(resource, property, contentHandler, lang);
                } catch (UnknownResourceException ure) {
                    System.err.println("Unknown property requested: " + property);
                }
            }
        } else {
            try {
                addProperty(resource, null, contentHandler, lang);
            } catch (UnknownResourceException ure) {
            }
        }
    }


    /**
     *  Adds a feature to the Property attribute of the VBSearch object
     *
     *@param  resource                      The feature to be added to the
     *      Property attribute
     *@param  property                      The feature to be added to the
     *      Property attribute
     *@param  lang                          The feature to be added to the
     *      Property attribute
     *@param  contentHandler                The feature to be added to the
     *      Property attribute
     *@exception  UnknownResourceException  Description of the Exception
     *@exception  SAXException              Description of the Exception
     */
    private void addProperty(VBSResource resource, String property, ContentHandler contentHandler, String lang) throws UnknownResourceException, SAXException {
        for (Iterator attriter = resource.listStatements(property);
                attriter.hasNext(); ) {
            VBSStatement statement = (VBSStatement) attriter.next();
            VBSResource presource = statement.getProperty();
            if (statement.getValueType() == VBSStatement.LITERAL) {
                VBSLiteral lvalue = statement.getLiteral();
                if (lvalue.matchLanguage(lang)) {
                    addPropertyElement(presource.getNamespace(), presource.getLocalName(),
                            getElementQName(presource), lvalue, contentHandler);
                }
            } else {
                VBSResource rvalue = statement.getResource();
                //System.out.println("addProperty" + resource + ", " + presource + ", " + rvalue);

                addPropertyElement(presource.getNamespace(), presource.getLocalName(),
                        getElementQName(presource), rvalue, contentHandler, true, lang);
            }
        }
    }


    /**
     *  Gets the attributeElement attribute of the VBSearch object
     *
     *@param  ns                Description of the Parameter
     *@param  resource          Description of the Parameter
     *@param  lang              Description of the Parameter
     *@param  incLabel          Description of the Parameter
     *@param  lname             The feature to be added to the PropertyElement
     *      attribute
     *@param  qname             The feature to be added to the PropertyElement
     *      attribute
     *@param  contentHandler    The feature to be added to the PropertyElement
     *      attribute
     *@exception  SAXException  Description of the Exception
     */
    private void addPropertyElement(String ns, String lname, String qname,
            VBSResource resource, ContentHandler contentHandler, boolean incLabel, String lang) throws SAXException {
        if (resource != null) {
            AttributesImpl attr = new AttributesImpl();
            if (resource.isAnon()) {
                if (resource instanceof Category) {
                    attr.addAttribute(VBSConf.OGT_NS, "catid",
                            this.prefixMapping.getNsURIPrefix(VBSConf.OGT_NS) + ":catid",
                            "CDATA", ((Category) resource).getCategoryID().toString(true));
                } else {
                    attr.addAttribute(VBSConf.RDF_NS, "nodeID",
                            this.prefixMapping.getNsURIPrefix(VBSConf.RDF_NS) + ":nodeID",
                            "CDATA", getNodeID(resource));
                }
            } else {
                attr.addAttribute(VBSConf.RDF_NS, "resource",
                        this.prefixMapping.getNsURIPrefix(VBSConf.RDF_NS) + ":resource",
                        "CDATA", resource.getURI());
            }
            if (incLabel) {
                //System.out.println(resource);
                // rdfs:label
                String label = resource.getLabel(lang);
                if (label != null) {
                    attr.addAttribute(VBSConf.RDFS_NS, "label", this.prefixMapping.getNsURIPrefix(VBSConf.RDFS_NS) + ":label",
                            "CDATA", label);
                }
            }
            contentHandler.startElement(ns, lname, qname, attr);
            contentHandler.endElement(ns, lname, qname);
        }
    }


    /**
     *  Gets the propertyElement attribute of the VBSearch object
     *
     *@param  ns                Description of the Parameter
     *@param  literal           Description of the Parameter
     *@param  lname             The feature to be added to the PropertyElement
     *      attribute
     *@param  qname             The feature to be added to the PropertyElement
     *      attribute
     *@param  contentHandler    The feature to be added to the PropertyElement
     *      attribute
     *@exception  SAXException  Description of the Exception
     */
    private void addPropertyElement(String ns, String lname, String qname,
            VBSLiteral literal, ContentHandler contentHandler) throws SAXException {
        if (literal != null) {
            AttributesImpl attr = new AttributesImpl();
            if (literal.getLanguage() != null) {
                attr.addAttribute("", "lang", "xml:lang", "CDATA", literal.getLanguage());
            }
            contentHandler.startElement(ns, lname, qname, attr);
            char[] ch = literal.getString().toCharArray();
            contentHandler.characters(ch, 0, ch.length);
            contentHandler.endElement(ns, lname, qname);
        }
    }


    /**
     *  Description of the Method
     *
     *@param  query             Description of the Parameter
     *@param  contentHandler    Description of the Parameter
     *@exception  SAXException  Description of the Exception
     */
    public void executeQuery(VBSQuery query, ContentHandler contentHandler) throws SAXException {
        executeQuery(query, new CachedQueryContext(this.index), contentHandler);
    }


    /**
     *  Description of the Method
     *
     *@param  query             Description of the Parameter
     *@param  context           Description of the Parameter
     *@param  contentHandler    Description of the Parameter
     *@exception  SAXException  Description of the Exception
     */
    public void executeQuery(VBSQuery query, QueryContext context, ContentHandler contentHandler) throws SAXException {
        contentHandler.startDocument();

        // rdf:RDF xmlns:...
        AttributesImpl nameSpaces = new AttributesImpl();
        for (Iterator prefixes = this.prefixMapping.getNsPrefixMap().entrySet().iterator(); prefixes.hasNext(); ) {
            Map.Entry prefixEntry = (Map.Entry) prefixes.next();
            String prefix = (String) prefixEntry.getKey();
            String uri = (String) prefixEntry.getValue();
            nameSpaces.addAttribute("", prefix, "xmlns:" + prefix, "CDATA", uri);
        }
        contentHandler.startElement(VBSConf.RDF_NS, "RDF", this.prefixMapping.getNsURIPrefix(VBSConf.RDF_NS) + ":RDF", nameSpaces);

        // ogt:CategoryBase
        VBSResource categoryBase = this.index.getCategoryBase();
        if (categoryBase != null) {
            startInstanceElement(categoryBase, VBSConf.OGT_NS, "CategoryBase",
                    this.prefixMapping.getNsURIPrefix(VBSConf.OGT_NS) + ":CategoryBase",
                    null, contentHandler, null);
            contentHandler.endElement(VBSConf.OGT_NS, "CategoryBase",
                    this.prefixMapping.getNsURIPrefix(VBSConf.OGT_NS) + ":CategoryBase");
        }

        // for each FacetSelector
        for (Iterator fsiter = query.getFacetSelectors(); fsiter.hasNext(); ) {
            FacetSelector fs = (FacetSelector) fsiter.next();
            //System.out.println(fs);

            Collection intersection = context.getIntersection(fs);
            String lang = fs.getAcceptLanguage();
            boolean incBookmarks = fs.getIncBookmarks();
            boolean incCategories = fs.getIncCategories();
            int incTopics = fs.getIncTopics();
            boolean incBookmarkHits = fs.getIncBookmarkHits();
            boolean incPathBookmarkHits = fs.getIncPathBookmarkHits();
            boolean incDirectBookmarkHits = fs.getIncDirectBookmarkHits();
            boolean incSubCategoryHits = fs.getIncSubCategoryHits();
            boolean incPathSubCategoryHits = fs.getIncPathSubCategoryHits();
            boolean incRemainder = fs.getIncRemainder();
            boolean incPosition = fs.getIncPosition();
            int maxSubCategories = -1;
            int maxBookmarks = -1;
            int skipSubCategories = 0;
            int skipBookmarks = 0;
            Map maxSubCategoriesMap = fs.getMaxSubCategories();
            Map maxBookmarksMap = fs.getMaxBookmarks();
            Map skipSubCategoriesMap = fs.getSkipSubCategories();
            Map skipBookmarksMap = fs.getSkipBookmarks();

            if (maxSubCategoriesMap != null && maxSubCategoriesMap.containsKey("")) {
                maxSubCategories = ((Integer) maxSubCategoriesMap.get("")).intValue();
            }
            if (maxBookmarksMap != null && maxBookmarksMap.containsKey("")) {
                maxBookmarks = ((Integer) maxBookmarksMap.get("")).intValue();
            }
            if (skipSubCategoriesMap != null && skipSubCategoriesMap.containsKey("")) {
                skipSubCategories = ((Integer) skipSubCategoriesMap.get("")).intValue();
            }
            if (skipBookmarksMap != null && skipBookmarksMap.containsKey("")) {
                skipBookmarks = ((Integer) skipBookmarksMap.get("")).intValue();
            }

            Collection groupedBookmarks = null;
            if (incBookmarks || incRemainder) {
                groupedBookmarks = new HashSet(intersection != null ? intersection.size() : 252);
            }
            Collection incCategoryProperties = null;
            Collection incBookmarkProperties = null;

            Iterator propiter = fs.getIncCategoryProperty();
            if (propiter != null) {
                incCategoryProperties = new LinkedList();
                while (propiter.hasNext()) {
                    incCategoryProperties.add(propiter.next());
                }
            }
            propiter = fs.getIncBookmarkProperty();
            if (propiter != null) {
                incBookmarkProperties = new LinkedList();
                while (propiter.hasNext()) {
                    incBookmarkProperties.add(propiter.next());
                }
            }

            contentHandler.startElement(VBSConf.OGT_NS, "Facet", this.prefixMapping.getNsURIPrefix(VBSConf.OGT_NS) + ":Facet", EMPTY_ATTRIBUTES);

            // ogt:facetSelector
            if (!fs.isAnon()) {
                addPropertyElement(VBSConf.OGT_NS, "facetSelector",
                        this.prefixMapping.getNsURIPrefix(VBSConf.OGT_NS) + ":facetSelector", fs, contentHandler, true, null);
            }

            // ogt:bookmarkHits
            if (incBookmarkHits) {
                int bookmarkHits = 0;
                if (intersection != null) {
                    bookmarkHits = intersection.size();
                } else {
                    bookmarkHits = this.index.getBookmarkCount();
                }
                contentHandler.startElement(VBSConf.OGT_NS, "bookmarkHits", this.prefixMapping.getNsURIPrefix(VBSConf.OGT_NS) + ":bookmarkHits", EMPTY_ATTRIBUTES);
                char[] ch = ("" + bookmarkHits).toCharArray();
                contentHandler.characters(ch, 0, ch.length);
                contentHandler.endElement(VBSConf.OGT_NS, "bookmarkHits", this.prefixMapping.getNsURIPrefix(VBSConf.OGT_NS) + ":bookmarkHits");
            }

            // ogt:subCategories
            if (incCategories || incSubCategoryHits) {
                Collection mc = context.getMatchingCategories(fs);
                if (mc == null) {
                    mc = new LinkedList();
                    for (Iterator i = this.index.getRoots(); i.hasNext(); ) {
                        mc.add(i.next());
                    }
                }
                Iterator subCategories = mc.iterator();
                handleSubCategories(subCategories,
                        incCategoryProperties,
                        incBookmarkProperties,
                        intersection,
                        incBookmarkHits,
                        incPathBookmarkHits,
                        incDirectBookmarkHits,
                        fs.getIncPath(),
                        fs.getIncSubCategories(),
                        fs.getFilter(),
                        incBookmarks,
                        incTopics,
                        (incCategories ? maxSubCategories : 0),
                        maxBookmarks,
                        maxSubCategoriesMap,
                        maxBookmarksMap,
                        incSubCategoryHits,
                        incPathSubCategoryHits,
                        skipSubCategories,
                        skipSubCategoriesMap,
                        skipBookmarksMap,
                        incPosition,
                        contentHandler,
                        lang,
                        groupedBookmarks);
            }

            // ogt:topicOf
            if (incBookmarks || incRemainder) {
                Collection matchingBookmarks = intersection;
                //System.out.println(matchingBookmarks);
                //System.out.println(intersection);
                if (matchingBookmarks == null) {
                    matchingBookmarks = this.index.getAllBookmarks();
                }
                int bmcount = 0;
                Iterator bmiter = matchingBookmarks.iterator();
                if (incBookmarks) {
                    contentHandler.startElement(VBSConf.OGT_NS, "topicOf", this.prefixMapping.getNsURIPrefix(VBSConf.OGT_NS) + ":topicOf", PARSE_TYPE_COLLECTION);
                    while (bmiter.hasNext() && (maxBookmarks < 0 || bmcount < (maxBookmarks + skipBookmarks))) {
                        Bookmark bookmark = (Bookmark) bmiter.next();
                        if (!groupedBookmarks.contains(bookmark)) {
                            if (bmcount >= skipBookmarks) {
                                startBookmarkElement(bookmark, fs, contentHandler);
                                if (incPosition) {
                                    contentHandler.startElement(VBSConf.OGT_NS, "position", this.prefixMapping.getNsURIPrefix(VBSConf.OGT_NS) + ":position", EMPTY_ATTRIBUTES);
                                    char[] ch = ("" + bmcount).toCharArray();
                                    contentHandler.characters(ch, 0, ch.length);
                                    contentHandler.endElement(VBSConf.OGT_NS, "position", this.prefixMapping.getNsURIPrefix(VBSConf.OGT_NS) + ":position");
                                }
                                contentHandler.endElement(VBSConf.BM_NS, "Bookmark", this.prefixMapping.getNsURIPrefix(VBSConf.BM_NS) + ":Bookmark");
                            }
                            bmcount++;
                        }
                    }
                    contentHandler.endElement(VBSConf.OGT_NS, "topicOf", this.prefixMapping.getNsURIPrefix(VBSConf.OGT_NS) + ":topicOf");
                }
                if (incRemainder) {
                    while (bmiter.hasNext()) {
                        if (!groupedBookmarks.contains(bmiter.next())) {
                            bmcount++;
                        }
                    }
                    contentHandler.startElement(VBSConf.OGT_NS, "remainder", this.prefixMapping.getNsURIPrefix(VBSConf.OGT_NS) + ":remainder", EMPTY_ATTRIBUTES);
                    char[] ch = ("" + bmcount).toCharArray();
                    contentHandler.characters(ch, 0, ch.length);
                    contentHandler.endElement(VBSConf.OGT_NS, "remainder", this.prefixMapping.getNsURIPrefix(VBSConf.OGT_NS) + ":remainder");
                }
            }
            contentHandler.endElement(VBSConf.OGT_NS, "Facet", this.prefixMapping.getNsURIPrefix(VBSConf.OGT_NS) + ":Facet");
        }
        contentHandler.endElement(VBSConf.RDF_NS, "RDF", this.prefixMapping.getNsURIPrefix(VBSConf.RDF_NS) + ":RDF");
        contentHandler.endDocument();
    }


    /**
     *@param  resource
     *@return
     */
    public String getElementQName(VBSResource resource) {
        String ns = resource.getNamespace();
        String nsprefix = (String) this.prefixMapping.getNsURIPrefix(ns);
        return nsprefix + ":" + resource.getLocalName();
    }


    /**
     *  Gets the index attribute of the VBSearch object
     *
     *@return    The index value
     */
    public VBSIndex getIndex() {
        return this.index;
    }


    /**
     *  Gets the nodeID attribute of the VBSearch object
     *
     *@param  resource  Description of the Parameter
     *@return           The nodeID value
     */
    public String getNodeID(VBSResource resource) {
        String key = resource.getID();
        String nodeID = (String) this.anonMap.get(key);
        if (nodeID == null) {
            nodeID = "A" + this.anonID++;
            this.anonMap.put(key, nodeID);
        }
        return nodeID;
    }


    /**
     *@param  subCategories
     *@param  incCategoryProperties
     *@param  incBookmarkProperties
     *@param  intersection
     *@param  incBookmarkHits
     *@param  incPath
     *@param  incSubCategories
     *@param  filter
     *@param  incBookmarks
     *@param  incTopics
     *@param  maxSubCategories
     *@param  maxBookmarks
     *@param  maxSubCategoriesMap
     *@param  maxBookmarksMap
     *@param  incSubCategoryHits
     *@param  skipSubCategories
     *@param  skipSubCategoriesMap
     *@param  skipBookmarksMap
     *@param  contentHandler
     *@param  lang
     *@param  groupedBookmarks
     *@param  incDirectBookmarkHits   Description of the Parameter
     *@param  incPathBookmarkHits     Description of the Parameter
     *@param  incPathSubCategoryHits  Description of the Parameter
     *@param  incPosition             Description of the Parameter
     *@return                         Description of the Return Value
     *@exception  SAXException
     */
    private boolean handleSubCategories(Iterator subCategories,
            Collection incCategoryProperties,
            Collection incBookmarkProperties,
            Collection intersection,
            boolean incBookmarkHits,
            boolean incPathBookmarkHits,
            boolean incDirectBookmarkHits,
            boolean incPath,
            int incSubCategories,
            int filter,
            boolean incBookmarks,
            int incTopics,
            int maxSubCategories,
            int maxBookmarks,
            Map maxSubCategoriesMap,
            Map maxBookmarksMap,
            boolean incSubCategoryHits,
            boolean incPathSubCategoryHits,
            int skipSubCategories,
            Map skipSubCategoriesMap,
            Map skipBookmarksMap,
            boolean incPosition,
            ContentHandler contentHandler,
            String lang,
            Collection groupedBookmarks) throws SAXException {
        int subCategoryHits = 0;
        boolean hasSubCategories = false;
        if (subCategories.hasNext()) {
            hasSubCategories = true;
            if (maxSubCategories != 0) {
                contentHandler.startElement(VBSConf.OGT_NS, "subCategories", this.prefixMapping.getNsURIPrefix(VBSConf.OGT_NS) + ":subCategories", PARSE_TYPE_COLLECTION);
                // shown sub categories
                while (subCategories.hasNext()
                         && (maxSubCategories < 0 || subCategoryHits < (maxSubCategories + skipSubCategories))) {
                    boolean skipped = subCategoryHits >= skipSubCategories;
                    boolean filtered = startCategoryElement((Category) subCategories.next(),
                            incCategoryProperties,
                            incBookmarkProperties,
                            intersection,
                            incBookmarkHits,
                            incPathBookmarkHits,
                            incDirectBookmarkHits,
                            incPath,
                            incSubCategories,
                            filter,
                            incBookmarks,
                            incTopics,
                            maxSubCategories,
                            maxBookmarks,
                            maxSubCategoriesMap,
                            maxBookmarksMap,
                            incSubCategoryHits,
                            incPathSubCategoryHits,
                            skipSubCategoriesMap,
                            skipBookmarksMap,
                            incPosition,
                            (skipped ? contentHandler : null),
                    // contentHandler --> Skipped?
                            lang,
                            groupedBookmarks);
                    if (!filtered) {
                        if (skipped) {
                            if (incPosition) {
                                contentHandler.startElement(VBSConf.OGT_NS, "position", this.prefixMapping.getNsURIPrefix(VBSConf.OGT_NS) + ":position", EMPTY_ATTRIBUTES);
                                char[] ch = ("" + subCategoryHits).toCharArray();
                                contentHandler.characters(ch, 0, ch.length);
                                contentHandler.endElement(VBSConf.OGT_NS, "position", this.prefixMapping.getNsURIPrefix(VBSConf.OGT_NS) + ":position");
                            }
                            contentHandler.endElement(VBSConf.OGT_NS, "Category", this.prefixMapping.getNsURIPrefix(VBSConf.OGT_NS) + ":Category");
                        }
                        subCategoryHits++;
                    }
                }
                contentHandler.endElement(VBSConf.OGT_NS, "subCategories", this.prefixMapping.getNsURIPrefix(VBSConf.OGT_NS) + ":subCategories");
            }
        }
        // Sub categories not shown
        if (incSubCategoryHits) {
            while (subCategories.hasNext()) {
                boolean filtered = startCategoryElement((Category) subCategories.next(),
                        incCategoryProperties,
                        incBookmarkProperties,
                        intersection,
                        false,
                // incBookmarkHits
                        false,
                // incPathBookmarkHits
                        false,
                // incDirectBookmarkHits
                        false,
                // incPath
                        FacetSelector.NONE,
                // incSubCategories
                        filter,
                        false,
                // incBookmarks
                        FacetSelector.NONE,
                // incTopics
                        0,
                // maxSubCategories
                        0,
                // maxBookmarks
                        null,
                // maxSubCategoriesMap
                        null,
                // maxBookmarksMap
                        false,
                // incSubCategoryHits
                        false,
                // incPathSubCategoryHits
                        null,
                // skipSubCategoriesMap
                        null,
                // skipBookmarksMap
                        false,
                // incPosition
                        null,
                // contentHandler = null --> silent
                        lang,
                        null);
                // groupedBookmarks
                if (!filtered) {
                    subCategoryHits++;
                }
            }
            contentHandler.startElement(VBSConf.OGT_NS, "subCategoryHits", this.prefixMapping.getNsURIPrefix(VBSConf.OGT_NS) + ":subCategoryHits", EMPTY_ATTRIBUTES);
            char[] ch = ("" + subCategoryHits).toCharArray();
            contentHandler.characters(ch, 0, ch.length);
            contentHandler.endElement(VBSConf.OGT_NS, "subCategoryHits", this.prefixMapping.getNsURIPrefix(VBSConf.OGT_NS) + ":subCategoryHits");
        }
        return hasSubCategories;
    }


    /**
     *  The main program for the VBSearch class
     *
     *@param  args  The command line arguments
     */
    public static void main(String[] args) {
        System.setErr(System.out);
        final String ERROR_MSG =
                "ERROR: --query and --input must be provided!\n" +
                "Classpath must include Jena2, and files ontogator.owl, bookmark.rdf and vbsearch.conf.\n" +
                "Usage: java -cp <classpath> vbsearch.VBSearch --input <file name> --query <file name> [options]\n" +
                "Options:\n" +
                "    --input          projection file (categories and bookmarks).\n" +
                "    --inputLang      language of input (default: RDF/XML).\n" +
                "    --query          name of the file containing the query.\n" +
                "    --queryLang      language of query (default: RDF/XML).\n" +
                "    --output         file name to store the results. If not\n" +
                "                     provided stdout is used.\n" +
                "    --repeat         number of repetitions.";
        try {
            VBSearch vbs = null;
            String inputLang = "RDF/XML";
            String queryLang = "RDF/XML";
            String query = "";
            int repeat = 1;
            
            boolean inputDefined = false;
            //OntModel input  = ModelFactory.createOntologyModel(ONT_MODEL_SPEC, null);
            HashMap options = new HashMap();
            for (int i = 0; i < args.length; i += 2) {
                options.put(args[i], args[i + 1]);
            }

            if (!options.containsKey("--query")) {
                System.out.println(ERROR_MSG);
                System.exit(0);
            }

            if (options.containsKey("--inputLang")) {
                inputLang = options.get("--inputLang").toString();
            }
            
            if (options.containsKey("--queryLang")) {
                queryLang = options.get("--queryLang").toString();
            }
            
            if (options.containsKey("--repeat")) {
                repeat = Integer.parseInt(options.get("--repeat").toString());
            }
            
            try {
                java.io.FileReader reader = new java.io.FileReader(options.get("--query").toString());
                int c = reader.read();
                while (c != -1) {
                    query += (char) c;
                    c = reader.read();
                }
                reader.close();
            } catch (Exception e) {
                e.printStackTrace();
            }
            
            System.out.println("********* VBSearch *********");
            System.out.print("Loading input... ");
            long time = System.currentTimeMillis();

            vbs = new VBSearch();

            if (options.containsKey("--input")) {
                OntModel input = null;
                input = ModelFactory.createOntologyModel(VBSConf.ONT_MODEL_SPEC, null);
                input.read(new FileInputStream(options.get("--input").toString()), VBSConf.BASE_URI, inputLang);
                System.out.println(System.currentTimeMillis() - time);

                System.out.print("Adding input... ");
                time = System.currentTimeMillis();
                vbs.add(input, "http://www.cs.helsinki.fi/u/ssaarela/vbstest.rdf");
                System.out.println(System.currentTimeMillis() - time);
                inputDefined = true;
            }

            if (!inputDefined) {
                System.out.println(ERROR_MSG);
                System.exit(0);
            }

            System.out.println(vbs.getIndex().toString());
            
            for (int i = 0; i < repeat; i++) {
                System.out.print("Executing query #"+(i+1)+"... ");
                time = System.currentTimeMillis();
                ByteArrayOutputStream results = new ByteArrayOutputStream();
                vbs.query(query, results);
                System.out.println(System.currentTimeMillis() - time);

                PrintStream writer = null;
                if (options.containsKey("--output")) {
                    writer = new PrintStream(new FileOutputStream(options.get("--output").toString()));
                } else {
                    writer = System.out;
                    writer.println("------ Results -------");
                }
                writer.println(results.toString("ISO-8859-1"));
                writer.flush();
                writer.close();
            }
        } catch (Exception e) {
            e.printStackTrace();
        }
    }


    /**
     *  Description of the Method
     *
     *@param  queryStr  Description of the Parameter
     *@param  results   Description of the Parameter
     */
    public void query(String queryStr, OutputStream results) {
        OntModel query = ModelFactory.createOntologyModel(VBSConf.ONT_MODEL_SPEC, null);
        //query.addSubModel(queryOntology);
        query((OntModel) query.read(new StringReader(queryStr), VBSConf.BASE_URI), results);
    }


    /**
     *  Description of the Method
     *
     *@param  query    Description of the Parameter
     *@param  results  Description of the Parameter
     */
    public void query(OntModel query, OutputStream results) {
        String rv = null;
        try {
            TransformerHandler handler = tfactory.newTransformerHandler();
            handler.setResult(new StreamResult(results));
            Transformer transformer = handler.getTransformer();
            transformer.setOutputProperty(OutputKeys.METHOD, "xml");
            transformer.setOutputProperty(OutputKeys.ENCODING, "UTF-8");
            executeQuery(new JenaVBSQuery(query), handler);
        } catch (SAXException saxe) {
            saxe.printStackTrace();
        } catch (javax.xml.transform.TransformerConfigurationException tce) {
            tce.printStackTrace();
        }
    }


    /**
     *  Gets the bookmarkElement attribute of the VBSearch object
     *
     *@param  bm                Description of the Parameter
     *@param  fs                Description of the Parameter
     *@param  contentHandler    The feature to be added to the BookmarkElement
     *      attribute
     *@exception  SAXException  Description of the Exception
     */
    public void startBookmarkElement(Bookmark bm, FacetSelector fs, ContentHandler contentHandler) throws SAXException {
        Collection incBookmarkProperties = null;
        Collection incCategoryProperties = null;

        Iterator iter = fs.getIncBookmarkProperty();
        if (iter != null) {
            incBookmarkProperties = new LinkedList();
            while (iter.hasNext()) {
                incBookmarkProperties.add(iter.next().toString());
            }
        }
        iter = fs.getIncCategoryProperty();
        if (iter != null) {
            incCategoryProperties = new LinkedList();
            while (iter.hasNext()) {
                incCategoryProperties.add(iter.next().toString());
            }
        }
        startBookmarkElement(bm,
                incBookmarkProperties,
                incCategoryProperties,
                fs.getIncBookmarkHits(),
                fs.getIncPathBookmarkHits(),
                fs.getIncTopics(),
                fs.getIncSubCategoryHits(),
                fs.getIncPathSubCategoryHits(),
                contentHandler,
                fs.getAcceptLanguage());
    }


    /**
     *  Gets the bookmarkElement attribute of the VBSearch object
     *
     *@param  bm                      Description of the Parameter
     *@param  incBookmarkProperties   Description of the Parameter
     *@param  incCategoryProperties   Description of the Parameter
     *@param  lang                    Description of the Parameter
     *@param  incTopics               Description of the Parameter
     *@param  incBookmarkHits         Description of the Parameter
     *@param  contentHandler          The feature to be added to the
     *      BookmarkElement attribute
     *@param  incSubCategoryHits
     *@param  incPathBookmarkHits     The feature to be added to the
     *      BookmarkElement attribute
     *@param  incPathSubCategoryHits  The feature to be added to the
     *      BookmarkElement attribute
     *@exception  SAXException        Description of the Exception
     */
    public void startBookmarkElement(Bookmark bm,
            Collection incBookmarkProperties,
            Collection incCategoryProperties,
            boolean incBookmarkHits,
            boolean incPathBookmarkHits,
            int incTopics,
            boolean incSubCategoryHits,
            boolean incPathSubCategoryHits,
            ContentHandler contentHandler,
            String lang) throws SAXException {
        startInstanceElement(bm, VBSConf.BM_NS, "Bookmark", this.prefixMapping.getNsURIPrefix(VBSConf.BM_NS) + ":Bookmark",
                incBookmarkProperties, contentHandler, lang);
        if (incTopics != FacetSelector.NONE) {
            contentHandler.startElement(VBSConf.BM_NS, "hasTopic", this.prefixMapping.getNsURIPrefix(VBSConf.BM_NS) + ":hasTopic", PARSE_TYPE_COLLECTION);
            for (Iterator topicsIter = bm.getCategories(); topicsIter.hasNext(); ) {
                startCategoryElement((Category) topicsIter.next(),
                        incCategoryProperties,
                        incBookmarkProperties,
                        null,
                // intersection
                        incBookmarkHits,
                        incPathBookmarkHits,
                        false,
                // incDirectBookmarkHits
                        (incTopics == FacetSelector.ALL ? true : false),
                        FacetSelector.NONE,
                        FacetSelector.NONE,
                        false,
                        FacetSelector.NONE,
                        -1,
                // maxSubCategories
                        -1,
                // maxBookmarks
                        null,
                // maxSubCategoriesMap
                        null,
                // maxBookmarksMap
                        incSubCategoryHits,
                        incPathSubCategoryHits,
                        null,
                // skipSubCategoriesMap
                        null,
                // skipBookmarksMap,
                        false,
                // incPosition
                        contentHandler,
                        lang,
                        null);
                contentHandler.endElement(VBSConf.OGT_NS, "Category", this.prefixMapping.getNsURIPrefix(VBSConf.OGT_NS) + ":Category");
            }
            contentHandler.endElement(VBSConf.BM_NS, "hasTopic", this.prefixMapping.getNsURIPrefix(VBSConf.BM_NS) + ":hasTopic");
        }
    }


    /**
     *  Description of the Method
     *
     *@param  category                Description of the Parameter
     *@param  incPath                 Description of the Parameter
     *@param  incSubCategories        Description of the Parameter
     *@param  incBookmarks            Description of the Parameter
     *@param  lang                    Description of the Parameter
     *@param  intersection            Description of the Parameter
     *@param  filter                  Description of the Parameter
     *@param  incBookmarkHits         Description of the Parameter
     *@param  incCategoryProperties   Description of the Parameter
     *@param  incBookmarkProperties   Description of the Parameter
     *@param  maxSubCategories
     *@param  maxBookmarks
     *@param  groupedBookmarks        Description of the Parameter
     *@param  incTopics               Description of the Parameter
     *@param  contentHandler          Description of the Parameter
     *@param  maxSubCategoriesMap
     *@param  maxBookmarksMap
     *@param  incSubCategoryHits
     *@param  skipSubCategoriesMap
     *@param  skipBookmarksMap
     *@param  incDirectBookmarkHits   Description of the Parameter
     *@param  incPathBookmarkHits     Description of the Parameter
     *@param  incPathSubCategoryHits  Description of the Parameter
     *@param  incPosition             Description of the Parameter
     *@return                         Description of the Return Value
     *@exception  SAXException        Description of the Exception
     */
    public boolean startCategoryElement(Category category,
            Collection incCategoryProperties,
            Collection incBookmarkProperties,
            Collection intersection,
            boolean incBookmarkHits,
            boolean incPathBookmarkHits,
            boolean incDirectBookmarkHits,
            boolean incPath,
            int incSubCategories,
            int filter,
            boolean incBookmarks,
            int incTopics,
            int maxSubCategories,
            int maxBookmarks,
            Map maxSubCategoriesMap,
            Map maxBookmarksMap,
            boolean incSubCategoryHits,
            boolean incPathSubCategoryHits,
            Map skipSubCategoriesMap,
            Map skipBookmarksMap,
            boolean incPosition,
            ContentHandler contentHandler,
            String lang,
            Collection groupedBookmarks) throws SAXException {
        //System.out.println("handle...");
        String catid = category.getCategoryID().toString(true);
        boolean filtered = true;
        int bookmarkHits = 0;
        boolean hasBookmarks = false;
        int skipSubCategories = 0;
        int skipBookmarks = 0;

        // Interpret limits
        if (maxSubCategoriesMap != null && maxSubCategoriesMap.containsKey(catid)) {
            maxSubCategories = ((Integer) maxSubCategoriesMap.get(catid)).intValue();
        }
        if (maxBookmarksMap != null && maxBookmarksMap.containsKey(catid)) {
            maxBookmarks = ((Integer) maxBookmarksMap.get(catid)).intValue();
        }
        if (skipSubCategoriesMap != null && skipSubCategoriesMap.containsKey(catid)) {
            skipSubCategories = ((Integer) skipSubCategoriesMap.get(catid)).intValue();
        }
        if (skipBookmarksMap != null && skipBookmarksMap.containsKey(catid)) {
            skipBookmarks = ((Integer) skipBookmarksMap.get(catid)).intValue();
        }

        if (incBookmarkHits || filter != FacetSelector.NONE || incBookmarks || groupedBookmarks != null) {
            for (Iterator bmiter = BasicIterator.uniqueIterator(category.getBookmarks(false)); bmiter.hasNext(); ) {
                //for (Iterator bmiter = category.getBookmarks(false); bmiter.hasNext(); ) {
                hasBookmarks = true;
                Object bookmark = bmiter.next();
                if (intersection == null || intersection.contains(bookmark)) {
                    //System.out.print(".");
                    bookmarkHits++;
                    if (groupedBookmarks != null) {
                        groupedBookmarks.add(bookmark);
                    } else if (!incBookmarkHits) {
                        // First hit --> no need to filter
                        break;
                    }
                }
            }
        }
        //System.out.println("Hits: " + bookmarkHits);
        // Apply desired filtering here --> return null if category at hand is filtered!!!
        if (filter == FacetSelector.NONE || (filter == FacetSelector.PRE_QUERY && hasBookmarks) || bookmarkHits > 0) {
            filtered = false;
            if (contentHandler != null) {
                // silent operation?
                // ogt:Category
                AttributesImpl attr = new AttributesImpl();
                attr.addAttribute(VBSConf.OGT_NS, "catid",
                        this.prefixMapping.getNsURIPrefix(VBSConf.OGT_NS) + ":catid",
                        "CDATA", "" + category.getCategoryID().toString(true));

                String label = category.getLabel(lang);
                if (label != null) {
                    attr.addAttribute(VBSConf.RDFS_NS, "label",
                            this.prefixMapping.getNsURIPrefix(VBSConf.RDFS_NS) + ":label",
                            "CDATA", label);
                }
                contentHandler.startElement(VBSConf.OGT_NS, "Category", this.prefixMapping.getNsURIPrefix(VBSConf.OGT_NS) + ":Category", attr);

                if (incBookmarkHits) {
                    contentHandler.startElement(VBSConf.OGT_NS, "bookmarkHits", this.prefixMapping.getNsURIPrefix(VBSConf.OGT_NS) + ":bookmarkHits", EMPTY_ATTRIBUTES);
                    char[] ch = ("" + bookmarkHits).toCharArray();
                    contentHandler.characters(ch, 0, ch.length);
                    contentHandler.endElement(VBSConf.OGT_NS, "bookmarkHits", this.prefixMapping.getNsURIPrefix(VBSConf.OGT_NS) + ":bookmarkHits");
                }

                addProperties(category, incCategoryProperties, contentHandler, lang);

                // ogt:hasRoot
                if (incPath) {
                    Category parent = category.getParent();
                    LinkedList path = null;
                    while (parent != null) {
                        if (path == null) {
                            path = new LinkedList();
                            addPropertyElement(VBSConf.OGT_NS, "subCategoryOf",
                                    this.prefixMapping.getNsURIPrefix(VBSConf.OGT_NS) + ":subCategoryOf", parent, contentHandler, true, lang);
                        }
                        path.addFirst(parent);
                        parent = parent.getParent();
                    }
                    if (path != null) {
                        contentHandler.startElement(VBSConf.OGT_NS, "hasRoot", this.prefixMapping.getNsURIPrefix(VBSConf.OGT_NS) + ":hasRoot", EMPTY_ATTRIBUTES);
                        addCategoryPath(path.iterator(),
                                incCategoryProperties,
                                intersection,
                                incPathBookmarkHits,
                                incPathSubCategoryHits,
                                contentHandler,
                                lang);
                        contentHandler.endElement(VBSConf.OGT_NS, "hasRoot", this.prefixMapping.getNsURIPrefix(VBSConf.OGT_NS) + ":hasRoot");
                    }
                }

                // ogt:subCategories
                if (incSubCategories != 0 || incSubCategoryHits) {
                    Iterator subCategories = category.getSubCategories(true);
                    handleSubCategories(subCategories,
                            incCategoryProperties,
                            incBookmarkProperties,
                            intersection,
                            incBookmarkHits,
                            incPathBookmarkHits,
                            incDirectBookmarkHits,
                            false,
                    // incPath
                            (incSubCategories > 0 ? incSubCategories - 1 : incSubCategories),
                            filter,
                            incBookmarks,
                            incTopics,
                            (incSubCategories != 0 ? maxSubCategories : 0),
                            maxBookmarks,
                            maxSubCategoriesMap,
                            maxBookmarksMap,
                            incSubCategoryHits,
                            incPathSubCategoryHits,
                            skipSubCategories,
                            skipSubCategoriesMap,
                            skipBookmarksMap,
                            incPosition,
                            contentHandler,
                            lang,
                            null
                    //groupedBookmarks
                            );
                }

                // ogt:topicOf
                if (incBookmarks || incDirectBookmarkHits) {
                    Iterator bmiter = null;
                    if (incSubCategories != 0) {
                        bmiter = category.getBookmarks(true);
                    } else if (incBookmarks) {
                        bmiter = BasicIterator.uniqueIterator(category.getBookmarks(false));
                    }
                    int bmcount = 0;
                    if (incBookmarks) {
                        contentHandler.startElement(VBSConf.OGT_NS, "topicOf", this.prefixMapping.getNsURIPrefix(VBSConf.OGT_NS) + ":topicOf", PARSE_TYPE_COLLECTION);
                        while (bmiter.hasNext() && (maxBookmarks < 0 || bmcount < (maxBookmarks + skipBookmarks))) {
                            Bookmark bookmark = (Bookmark) bmiter.next();
                            if (intersection == null || intersection.contains(bookmark)) {
                                if (bmcount >= skipBookmarks) {
                                    startBookmarkElement(bookmark,
                                            incBookmarkProperties,
                                            incCategoryProperties,
                                            incBookmarkHits,
                                            incPathBookmarkHits,
                                            incTopics,
                                            incSubCategoryHits,
                                            incPathSubCategoryHits,
                                            contentHandler,
                                            lang);
                                    if (incPosition) {
                                        contentHandler.startElement(VBSConf.OGT_NS, "position", this.prefixMapping.getNsURIPrefix(VBSConf.OGT_NS) + ":position", EMPTY_ATTRIBUTES);
                                        char[] ch = ("" + bmcount).toCharArray();
                                        contentHandler.characters(ch, 0, ch.length);
                                        contentHandler.endElement(VBSConf.OGT_NS, "position", this.prefixMapping.getNsURIPrefix(VBSConf.OGT_NS) + ":position");
                                    }
                                    contentHandler.endElement(VBSConf.BM_NS, "Bookmark", this.prefixMapping.getNsURIPrefix(VBSConf.BM_NS) + ":Bookmark");
                                }
                                bmcount++;
                            }
                        }
                        contentHandler.endElement(VBSConf.OGT_NS, "topicOf", this.prefixMapping.getNsURIPrefix(VBSConf.OGT_NS) + ":topicOf");
                    }
                    if (incDirectBookmarkHits && incSubCategories != 0) {
                        while (bmiter.hasNext()) {
                            Bookmark bookmark = (Bookmark) bmiter.next();
                            if (intersection == null || intersection.contains(bookmark)) {
                                bmcount++;
                            }
                        }
                        contentHandler.startElement(VBSConf.OGT_NS, "directBookmarkHits", this.prefixMapping.getNsURIPrefix(VBSConf.OGT_NS) + ":directBookmarkHits", EMPTY_ATTRIBUTES);
                        char[] ch = ("" + bmcount).toCharArray();
                        contentHandler.characters(ch, 0, ch.length);
                        contentHandler.endElement(VBSConf.OGT_NS, "directBookmarkHits", this.prefixMapping.getNsURIPrefix(VBSConf.OGT_NS) + ":directBookmarkHits");
                    }
                }
            }
        }
        return filtered;
    }


    /**
     *  Gets the instanceElement attribute of the VBSearch object
     *
     *@param  ns                Description of the Parameter
     *@param  resource          Description of the Parameter
     *@param  lang              Description of the Parameter
     *@param  incProperties     Description of the Parameter
     *@param  lname             Description of the Parameter
     *@param  qname             Description of the Parameter
     *@param  contentHandler    Description of the Parameter
     *@exception  SAXException  Description of the Exception
     */
    private void startInstanceElement(VBSResource resource, String ns, String lname, String qname,
            Collection incProperties, ContentHandler contentHandler, String lang) throws SAXException {
        AttributesImpl attr = new AttributesImpl();
        // rdf:nodeID - rdf:about
        if (resource.isAnon()) {
            attr.addAttribute(VBSConf.RDF_NS, "nodeID",
                    this.prefixMapping.getNsURIPrefix(VBSConf.RDF_NS) + ":nodeID",
                    "CDATA", getNodeID(resource));
        } else {
            attr.addAttribute(VBSConf.RDF_NS, "about",
                    this.prefixMapping.getNsURIPrefix(VBSConf.RDF_NS) + ":about",
                    "CDATA", resource.getURI());
        }
        // rdfs:label
        String label = resource.getLabel(lang);
        if (label != null) {
            attr.addAttribute(VBSConf.RDFS_NS, "label",
                    this.prefixMapping.getNsURIPrefix(VBSConf.RDFS_NS) + ":label",
                    "CDATA", label);
        }
        contentHandler.startElement(ns, lname, qname, attr);
        addProperties(resource, incProperties, contentHandler, lang);
    }
}

