/*
 *  Copyright (c) 2003 Helsinki Institute for Information Technology
 *
 *  Permission is hereby granted, free of charge, to any person obtaining a copy
 *  of this software and associated documentation files (the "Software"), to
 *  deal in the Software without restriction, including without limitation the
 *  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 *  sell copies of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *
 *  The above copyright notice and this permission notice shall be included in
 *  all copies or substantial portions of the Software.
 *
 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 *  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 *  IN THE SOFTWARE.
 */

package vbsearch.impl.mem;

import java.io.Serializable;
import java.net.*;
import java.util.*;
import java.util.regex.*;
import seco.util.iterator.*;
import vbsearch.*;
import vbsearch.impl.*;
import vbsearch.util.*;

/**
 *  Description of the Class
 *
 *@author     Samppa
 *@created    26. elokuuta 2003
 */
public class VBSIndexImpl implements VBSIndex, Serializable {
    /**
     *  Description of the Field
     */
    private Pattern SENTENCE_SPLIT_PATTERN = Pattern.compile("[\\.\\,\\:\\;\\?\\!]");

    /**
     *  Description of the Field
     */
    private Pattern WORD_SPLIT_PATTERN = Pattern.compile("\\s");

    /**
     *  Description of the Field
     */
    private Map bmidmap = new HashMap();
    /**
     *  Description of the Field
     */
    private SortedMap bmindex = new TreeMap();
    /**
     *  Description of the Field
     */
    private SortedMap bookmarkLiteralIndex = new TreeMap();

    private VBSResource categoryBase = null;

    /**
     *  Description of the Field
     */
    private SortedMap categoryLiteralIndex = new TreeMap();
    /**
     *  Categories by their URIs/IDs.
     */
    private Map cidmap = new HashMap();
    /**
     *  Categories
     */
    private SortedMap cindex = new TreeMap();
    /**
     *  Description of the Field
     */
    private String[] literalBookmarkIndexProperties = VBSConf.LITERAL_BOOKMARK_INDEX_PROPERTIES;
    /**
     *  Description of the Field
     */
    private String[] literalCategoryIndexProperties = VBSConf.LITERAL_CATEGORY_INDEX_PROPERTIES;
    /**
     *  Description of the Field
     */
    private PrefixLabel nextRoot = PrefixLabel.getRoot();
    /**
     *  Description of the Field
     */
    private Map projectionOfIndex = new HashMap();

    /**
     *  Description of the Field
     */
    private Collection roots = new LinkedList();


    /**
     *  Adds a feature to the Bookmark attribute of the VBSIndex object
     *
     *@param  bminfo                        The feature to be added to the
     *      Bookmark attribute
     *@return                               Description of the Return Value
     *@exception  NullPointerException      Description of the Exception
     *@exception  UnknownResourceException  Description of the Exception
     *@exception  URISyntaxException        Description of the Exception
     */
    public Bookmark addBookmark(Bookmark bminfo) throws NullPointerException, UnknownResourceException, URISyntaxException {
        Bookmark bm = null;
        Iterator topics = bminfo.getCategories();

        // Add bookmark for all topics
        if (topics != null) {
            while (topics.hasNext()) {
                bm = addBookmark((CategoryImpl) getValue(this.cidmap, (Category) topics.next()), bminfo);
            }
        }
        return bm;
    }


    /**
     *  Adds a feature to the Bookmark attribute of the VBSIndexImpl object
     *
     *@param  category                      The feature to be added to the
     *      Bookmark attribute
     *@param  bminfo                        The feature to be added to the
     *      Bookmark attribute
     *@return                               Description of the Return Value
     *@exception  NullPointerException      Description of the Exception
     *@exception  URISyntaxException        Description of the Exception
     *@exception  UnknownResourceException  Description of the Exception
     */
    protected Bookmark addBookmark(CategoryImpl category, Bookmark bminfo) throws NullPointerException, URISyntaxException, UnknownResourceException {
        BookmarkImpl bm = null;
        //System.out.println("Adding bookmark " + bminfo + "...");

        // Is bookmark known already?
        if (hasValue(this.bmidmap, bminfo)) {
            bm = (BookmarkImpl) getValue(this.bmidmap, bminfo);
        } else {
            bm = new BookmarkImpl(bminfo);
            // Add id map entry
            addValue(this.bmidmap, bm);
        }

        // Establish double linking between category and bookmark
        if (!category.hasBookmark(bm)) {
            category.addBookmark(bm);
        }
        if (!bm.hasCategory(category)) {
            PrefixLabel catid = getNextSubLabel(category);
            bm.addCategory(category, catid);
            // Add index entry for the new category id of bookmark
            this.bmindex.put(catid, bm);
        }
        // Add literal index entries for bookmark
        if (literalBookmarkIndexProperties != null) {
            for (int i = 0; i < literalBookmarkIndexProperties.length; i++) {
                for (Iterator labeliter = bm.listStatements(literalBookmarkIndexProperties[i]);
                        labeliter.hasNext(); ) {
                    addLiteralIndexEntries(bm,
                            ((VBSStatement) labeliter.next()).getLiteral().getString(),
                            this.bookmarkLiteralIndex);
                }
            }
        }
        return bm;
    }


    /**
     *  Adds a feature to the Category attribute of the VBSIndex object
     *
     *@param  catinfo                       The feature to be added to the
     *      Category attribute
     *@return                               Description of the Return Value
     *@exception  NullPointerException      Description of the Exception
     *@exception  UnknownResourceException  Description of the Exception
     *@exception  URISyntaxException        Description of the Exception
     */
    public Category addCategory(Category catinfo) throws NullPointerException, UnknownResourceException, URISyntaxException {
        return addCategory((CategoryImpl) getValue(this.cidmap, catinfo.getParent()), catinfo);
    }


    /**
     *  Adds a feature to the Category attribute of the VBSIndexImpl object
     *
     *@param  parent                        The feature to be added to the
     *      Category attribute
     *@param  catinfo                       The feature to be added to the
     *      Category attribute
     *@return                               Description of the Return Value
     *@exception  NullPointerException      Description of the Exception
     *@exception  URISyntaxException        Description of the Exception
     *@exception  UnknownResourceException  Description of the Exception
     */
    protected Category addCategory(CategoryImpl parent, Category catinfo) throws NullPointerException, UnknownResourceException, URISyntaxException {
        CategoryImpl category = null;
        // Check if category exists in the index already
        if (hasValue(this.cidmap, catinfo)) {
            category = (CategoryImpl) getValue(this.cidmap, catinfo);
        } else {
            category = new CategoryImpl(parent, catinfo, this);

            // Add id map entry
            addValue(this.cidmap, category);
            //this.cidmap.put((catinfo.isAnon() ? catinfo.getID() : catinfo.getURI()), category);

            //  Set category id
            category.setCategoryID(getNextSubLabel(parent));

            // Link to super category or roots
            if (parent != null) {
                if (!parent.hasSubCategory(category)) {
                    parent.addSubCategory(category);
                }
            } else {
                this.roots.add(category);
            }

            // Add index entry for catid
            this.cindex.put(category.getCategoryID(), category);

            // Add index entries for ogt:projectionOf
            for (Iterator projiter = category.listStatements(VBSConf.OGT_NS + "projectionOf");
                    projiter.hasNext(); ) {
                VBSResource projectionOf = ((VBSStatement) projiter.next()).getResource();
                //System.out.println("Projection index entry: " + projectionOf);
                addMultiValue(this.projectionOfIndex,
                        (projectionOf.isAnon() ? projectionOf.getID() : projectionOf.getURI()),
                        category);
            }

            // Add literal index entries for category
            if (literalCategoryIndexProperties != null) {
                for (int i = 0; i < literalCategoryIndexProperties.length; i++) {
                    for (Iterator labeliter = category.listStatements(literalCategoryIndexProperties[i]);
                            labeliter.hasNext(); ) {
                        addLiteralIndexEntries(category,
                                ((VBSStatement) labeliter.next()).getLiteral().getString(),
                                this.categoryLiteralIndex);
                    }
                }
            }

            // recurse over sub categories
            Iterator subCategories = catinfo.getSubCategories(true);
            if (subCategories != null) {
                while (subCategories.hasNext()) {
                    addCategory(category, (Category) subCategories.next());
                }
            }

            // add bookmarks
            Iterator bookmarks = catinfo.getBookmarks(true);
            if (bookmarks != null) {
                while (bookmarks.hasNext()) {
                    addBookmark(category, (Bookmark) bookmarks.next());
                }
            }
        }
        return category;
    }


    /**
     *  Adds a feature to the LiteralIndexEntries attribute of the VBSIndexImpl
     *  object
     *
     *@param  s        The feature to be added to the LiteralIndexEntries
     *      attribute
     *@param  literal  The feature to be added to the LiteralIndexEntries
     *      attribute
     *@param  index    The feature to be added to the LiteralIndexEntries
     *      attribute
     */
    protected void addLiteralIndexEntries(Object s, String literal, Map index) {
        //System.out.println("Text: " + literal);
        String[] literalSplit = SENTENCE_SPLIT_PATTERN.split(literal);
        for (int i = 0; i < literalSplit.length; i++) {
            String sentence = literalSplit[i].trim();
            //System.out.println(" Sentence: " + sentence);
            Matcher m = WORD_SPLIT_PATTERN.matcher(sentence);
            int start = 0;
            int len = sentence.length();
            do {
                //System.out.println("  Suffix: " + sentence.substring(start, len));
                addMultiValue(index, new PrefixLabel(sentence.substring(start, len).toUpperCase(), false), s);
                if (m.find(start)) {
                    start = m.end();
                } else {
                    start = -1;
                }
            } while (start > 0 && start < len);
        }
    }


    /**
     *  Adds a feature to the Value attribute of the VBSIndexImpl class
     *
     *@param  map    The feature to be added to the Value attribute
     *@param  key    The feature to be added to the Value attribute
     *@param  value  The feature to be added to the Value attribute
     */
    protected static void addMultiValue(Map map, Object key, Object value) {
        Object o = map.get(key);
        if (o == null) {
            o = value;
        } else {
            Collection c;
            if (o instanceof Collection) {
                c = (Collection) o;
            } else {
                c = new LinkedList();
                c.add(o);
            }
            c.add(value);
            o = c;
        }
        map.put(key, o);
    }


    /**
     *  Adds a feature to the Value attribute of the VBSIndexImpl class
     *
     *@param  map       The feature to be added to the Value attribute
     *@param  resource  The feature to be added to the Value attribute
     */
    protected static void addValue(Map map, VBSResource resource) {
        if (resource == null) {
            return;
        }
        String key = (resource.isAnon() ? resource.getID() : resource.getURI());
        map.put(key, resource);
    }


    /**
     *  Gets the allBookmarks attribute of the VBSIndexImpl object
     *
     *@return    The allBookmarks value
     */
    public Collection getAllBookmarks() {
        return this.bmidmap.values();
    }


    /**
     *  Gets the bookmark attribute of the VBSIndex object
     *
     *@param  catid  Description of the Parameter
     *@return        The bookmark value
     */
    public Bookmark getBookmark(String catid) {
        return getBookmark(new PrefixLabel(catid, true));
    }


    /**
     *  Gets the bookmark attribute of the VBSIndex object
     *
     *@param  catid  Description of the Parameter
     *@return        The bookmark value
     */
    public Bookmark getBookmark(PrefixLabel catid) {
        return (Bookmark) this.bmindex.get(catid);
    }


    /**
     *@param  uri
     *@return
     */
    public Bookmark getBookmarkByURI(String uri) {
        return (Bookmark) this.bmidmap.get(uri);
    }


    /**
     *  Gets the bookmarkCount attribute of the VBSIndexImpl object
     *
     *@return    The bookmarkCount value
     */
    public int getBookmarkCount() {
        return this.bmidmap.size();
    }


    /**
     *  Gets the bookmarks attribute of the VBSIndex object
     *
     *@param  category  Description of the Parameter
     *@return           The bookmarks value
     */
    protected Collection getBookmarks(Category category) {
        PrefixLabel catid = category.getCategoryID();
        return this.bmindex.subMap(catid, catid.getNext()).values();
    }


    /**
     *  Gets the bookmarksByKeyword attribute of the VBSIndexImpl object
     *
     *@param  keyword  Description of the Parameter
     *@return          The bookmarksByKeyword value
     */
    public Iterator getBookmarksByKeyword(String keyword) {
        Iterator rv = null;
        PrefixLabel kw = new PrefixLabel(keyword.toUpperCase(), false);
        rv = getMultiValues(this.bookmarkLiteralIndex.subMap(kw, kw.getNext()), null);
        return rv;
    }


    /**
     *  Gets the keywordMatches attribute of the VBSIndexImpl object
     *
     *@param  categories  Description of the Parameter
     *@param  keyword     Description of the Parameter
     *@return             The keywordMatches value
     */
    public Iterator getCategoriesByKeyword(String keyword, final Collection categories) {
        Iterator rv = null;
        PrefixLabel kw = new PrefixLabel(keyword.toUpperCase(), false);
        rv = getMultiValues(this.categoryLiteralIndex.subMap(kw, kw.getNext()), null);
        if (categories != null) {
            rv = new BasicIterator(rv,
                new IterationFilter() {
                    public boolean accept(Object o) {
                        Category category = (Category) o;
                        for (Iterator catiter = categories.iterator(); catiter.hasNext(); ) {
                            if (isSubCategoryOf(category, (Category) catiter.next())) {
                                return true;
                            }
                        }
                        return false;
                    }
                });
        }
        return rv;
    }


    /**
     *  Gets the categoriesByProjection attribute of the VBSIndexImpl object
     *
     *@param  projectedFrom  Description of the Parameter
     *@param  categories     Description of the Parameter
     *@return                The categoriesByProjection value
     */
    public Iterator getCategoriesByProjection(String projectedFrom, final Collection categories) {
        Iterator rv = null;
        rv = getMultiValues(this.projectionOfIndex, projectedFrom);
        if (categories != null) {
            rv = new BasicIterator(rv,
                new IterationFilter() {
                    public boolean accept(Object o) {
                        Category category = (Category) o;
                        for (Iterator catiter = categories.iterator(); catiter.hasNext(); ) {
                            if (isSubCategoryOf(category, (Category) catiter.next())) {
                                return true;
                            }
                        }
                        return false;
                    }
                });
        }
        return rv;
    }


    /**
     *  Gets the category attribute of the VBSIndex object
     *
     *@param  catid  Description of the Parameter
     *@return        The category value
     */
    public Category getCategory(String catid) {
        return getCategory(new PrefixLabel(catid, true));
    }


    /**
     *  Gets the category attribute of the VBSIndex object
     *
     *@param  catid  Description of the Parameter
     *@return        The category value
     */
    public Category getCategory(PrefixLabel catid) {
        return (Category) this.cindex.get(catid);
    }


    /**
     *  Gets the categoryBase attribute of the VBSIndexImpl object
     *
     *@return    The categoryBase value
     */
    public VBSResource getCategoryBase() {
        return this.categoryBase;
    }


    /**
     *  Gets the values attribute of the VBSIndexImpl class
     *
     *@param  map  Description of the Parameter
     *@param  key  Description of the Parameter
     *@return      The values value
     */
    protected static Iterator getMultiValues(Map map, Object key) {
        BasicIterator rv = null;
        if (key == null) {
            // return all values
            Iterator values = map.values().iterator();
            // First value or collection...
            if (values.hasNext()) {
                Object o = values.next();
                if (o instanceof Collection) {
                    rv = new BasicIterator(((Collection) o).iterator());
                } else {
                    rv = new BasicIterator(o);
                }
            }
            // ... and rest of values/collections
            while (values.hasNext()) {
                Object o = values.next();
                if (o instanceof Collection) {
                    rv = rv.andThen(((Collection) o).iterator());
                } else {
                    rv = rv.andThen(new BasicIterator(o));
                }
            }
        } else {
            // return values of given key
            Object o = map.get(key);
            if (o != null) {
                if (o instanceof Collection) {
                    rv = new BasicIterator(((Collection) o).iterator());
                } else {
                    rv = new BasicIterator(o);
                }
            }
        }
        if (rv == null) {
            rv = new BasicIterator();
        }
        return rv;
    }


    /**
     *  Gets the subLabel attribute of the VBSIndexImpl object
     *
     *@param  parent  Description of the Parameter
     *@return         The subLabel value
     */
    private PrefixLabel getNextSubLabel(Category parent) {
        PrefixLabel rv = null;
        if (parent == null) {
            rv = this.nextRoot;
            this.nextRoot = this.nextRoot.getNext();
        } else {
            rv = parent.getCategoryID().getNextSub();
        }
        return rv;
    }


    /**
     *  Gets the roots attribute of the VBSIndexImpl object
     *
     *@return    The roots value
     */
    public Iterator getRoots() {
        return this.roots.iterator();
    }


    /**
     *  Gets the subCategories attribute of the VBSIndex object
     *
     *@param  category  Description of the Parameter
     *@return           The subCategories value
     */
    protected Collection getSubCategories(Category category) {
        PrefixLabel catid = category.getCategoryID();
        return this.cindex.subMap(catid, catid.getNext()).values();
    }


    /**
     *  Gets the value attribute of the VBSIndexImpl class
     *
     *@param  map                           Description of the Parameter
     *@param  resource                      Description of the Parameter
     *@return                               The value value
     *@exception  UnknownResourceException  Description of the Exception
     */
    protected static VBSResource getValue(Map map, VBSResource resource) throws UnknownResourceException {
        if (resource == null) {
            return null;
        }
        String key = (resource.isAnon() ? resource.getID() : resource.getURI());
        VBSResource rv = (VBSResource) map.get(key);
        if (rv == null) {
            throw new UnknownResourceException(resource);
        }
        return rv;
    }


    /**
     *  Description of the Method
     *
     *@param  map       Description of the Parameter
     *@param  resource  Description of the Parameter
     *@return           Description of the Return Value
     */
    protected static boolean hasValue(Map map, VBSResource resource) {
        boolean rv = false;
        if (resource != null) {
            String key = (resource.isAnon() ? resource.getID() : resource.getURI());
            rv = map.containsKey(key);
        }
        return rv;
    }


    /**
     *  Gets the subCategoryOf attribute of the VBSIndexImpl object
     *
     *@param  child   Description of the Parameter
     *@param  parent  Description of the Parameter
     *@return         The subCategoryOf value
     */
    protected boolean isSubCategoryOf(Category child, Category parent) {
        boolean rv = false;
        PrefixLabel ccatid = child.getCategoryID();
        PrefixLabel pcatid = parent.getCategoryID();
        if (ccatid != null && pcatid != null) {
            if (ccatid.getString().startsWith(pcatid.getString())) {
                rv = true;
            }
        }
        return rv;
    }


    /**
     *  Description of the Method
     *
     *@param  bminfo                        Description of the Parameter
     *@exception  UnknownResourceException  Description of the Exception
     */
    public void removeBookmark(Bookmark bminfo) throws UnknownResourceException {
        BookmarkImpl bm = (BookmarkImpl) getValue(this.bmidmap, bminfo);

        // Remove bookmark from all of it's topics
        Iterator topics = bm.getCategories();
        if (topics != null) {
            while (topics.hasNext()) {
                removeBookmark((CategoryImpl) topics.next(), bm);
            }
        }
        if (literalBookmarkIndexProperties != null) {
            for (int i = 0; i < literalBookmarkIndexProperties.length; i++) {
                for (Iterator labeliter = bm.listStatements(literalBookmarkIndexProperties[i]);
                        labeliter.hasNext(); ) {
                    removeLiteralIndexEntries(bm,
                            ((VBSStatement) labeliter.next()).getLiteral().getString(),
                            this.bookmarkLiteralIndex);
                }
            }
        }
    }


    /**
     *  Description of the Method
     *
     *@param  bminfo                        Description of the Parameter
     *@param  category                      Description of the Parameter
     *@exception  UnknownResourceException  Description of the Exception
     */
    protected void removeBookmark(CategoryImpl category, Bookmark bminfo) throws UnknownResourceException {
        BookmarkImpl bm = (BookmarkImpl) getValue(this.bmidmap, bminfo);
        //System.out.println("Remove bookmark " + bm + " of category " + category);
        // Remove linking between category and bookmark
        category.removeBookmark(bm);
        PrefixLabel catid = bm.removeCategory(category);
        this.bmindex.remove(catid);
        if (!bm.hasCategories()) {
            removeValue(this.bmidmap, bm);
        }
    }


    /**
     *  Description of the Method
     *
     *@param  catinfo                       Description of the Parameter
     *@exception  UnknownResourceException  Description of the Exception
     *@exception  NullPointerException      Description of the Exception
     */
    public void removeCategory(Category catinfo) throws UnknownResourceException, NullPointerException {
        removeCategory((CategoryImpl) getValue(this.cidmap, catinfo.getParent()), catinfo);
    }


    /**
     *  Description of the Method
     *
     *@param  catinfo                       Description of the Parameter
     *@param  parent                        Description of the Parameter
     *@exception  UnknownResourceException  Description of the Exception
     *@exception  NullPointerException      Description of the Exception
     */
    protected void removeCategory(CategoryImpl parent, Category catinfo) throws UnknownResourceException, NullPointerException {
        CategoryImpl category = (CategoryImpl) getValue(this.cidmap, catinfo);
        //System.out.println(".");
        System.out.println("Remove category " + category);
        // remove super category or root link
        if (parent == null) {
            this.roots.remove(category);
        } else {
            parent.removeSubCategory(category);
        }
        // remove id map entry
        removeValue(this.cidmap, category);
        //this.cidmap.remove((category.isAnon() ? category.getID() : category.getURI()));

        // remove index entry
        this.cindex.remove(category.getCategoryID());

        // remove ogt:projectionOf index entries
        for (Iterator projiter = category.listStatements(VBSConf.OGT_NS + "projectionOf");
                projiter.hasNext(); ) {
            VBSResource projectionOf = ((VBSStatement) projiter.next()).getResource();
            //System.out.println("Projection index entry: " + projectionOf);
            removeMultiValue(this.projectionOfIndex,
                    (projectionOf.isAnon() ? projectionOf.getID() : projectionOf.getURI()),
                    category);
        }

        // remove index entries for rdfs:label
        if (literalCategoryIndexProperties != null) {
            for (int i = 0; i < literalCategoryIndexProperties.length; i++) {
                for (Iterator labeliter = category.listStatements(literalCategoryIndexProperties[i]);
                        labeliter.hasNext(); ) {
                    removeLiteralIndexEntries(category,
                            ((VBSStatement) labeliter.next()).getLiteral().getString(),
                            this.categoryLiteralIndex);
                }
            }
        }

        // remove bookmarks
        Iterator bookmarks = category.getBookmarks(true);
        if (bookmarks != null) {
            // Two phase removal to avoid ConcurrentModificationException
            LinkedList removeList = new LinkedList();
            while (bookmarks.hasNext()) {
                removeList.add(bookmarks.next());
            }
            for (Iterator i = removeList.iterator(); i.hasNext(); ) {
                removeBookmark(category, (Bookmark) i.next());
            }
        }

        // remove sub categories recursively
        Iterator subCategories = category.getSubCategories(true);
        if (subCategories != null) {
            // Two phase removal to avoid ConcurrentModificationException
            LinkedList removeList = new LinkedList();
            while (subCategories.hasNext()) {
                removeList.add(subCategories.next());
            }
            //System.out.println("subcategories: " + removeList.size());
            for (Iterator i = removeList.iterator(); i.hasNext(); ) {
                removeCategory(category, (Category) i.next());
            }
        }
    }


    /**
     *  Description of the Method
     *
     *@param  s        Description of the Parameter
     *@param  literal  Description of the Parameter
     *@param  index    Description of the Parameter
     */
    protected void removeLiteralIndexEntries(Object s, String literal, Map index) {
        String[] literalSplit = SENTENCE_SPLIT_PATTERN.split(literal);
        for (int i = 0; i < literalSplit.length; i++) {
            String sentence = literalSplit[i].trim();
            Matcher m = WORD_SPLIT_PATTERN.matcher(sentence);
            int start = 0;
            int len = sentence.length();
            do {
                removeMultiValue(index, new PrefixLabel(sentence.substring(start, len).toUpperCase(), false), s);
                if (m.find(start)) {
                    start = m.end();
                } else {
                    start = -1;
                }
            } while (start > 0 && start < len);
        }
    }


    /**
     *  Description of the Method
     *
     *@param  map    Description of the Parameter
     *@param  key    Description of the Parameter
     *@param  value  Description of the Parameter
     */
    protected static void removeMultiValue(Map map, Object key, Object value) {
        Object o = map.get(key);
        if (o != null) {
            if (o instanceof Collection) {
                Collection c = (Collection) o;
                c.remove(value);
                if (c.size() < 2) {
                    o = null;
                    Iterator i = c.iterator();
                    if (i.hasNext()) {
                        o = i.next();
                    }
                    map.remove(key);
                    if (o != null) {
                        map.put(key, o);
                    }
                }
            } else {
                map.remove(key);
            }
        }
    }


    /**
     *  Description of the Method
     *
     *@param  map       Description of the Parameter
     *@param  resource  Description of the Parameter
     */
    protected static void removeValue(Map map, VBSResource resource) {
        if (resource == null) {
            return;
        }
        String key = (resource.isAnon() ? resource.getID() : resource.getURI());
        map.remove(key);
    }


    /**
     *  Sets the categoryBase attribute of the VBSIndexImpl object
     *
     *@param  cb  The new categoryBase value
     */
    public void setCategoryBase(VBSResource cb) {
        this.categoryBase = cb;
    }


    /**
     *  Description of the Method
     *
     *@return    Description of the Return Value
     */
    public String toString() {
        return  "bookmarks                      = " + this.bmidmap.size() + "\n" +
                "categories                     = " + this.cidmap.size() + "\n" +
                "roots                          = " + roots.size() + "\n" +
                "category-bookmark -relations   = " + this.bmindex.size() + "\n" +
                "projected resources            = " + this.projectionOfIndex.size() + "\n" +
                "category keywords              = " + this.categoryLiteralIndex.size() + "\n" +
                "bookmark keywords              = " + this.bookmarkLiteralIndex.size() + "\n" +
                "avg bookmarks/category         = " + ((double)this.bmindex.size() / (double)this.cidmap.size()) + "\n" +
                "avg topics/bookmark             = " + ((double)this.bmindex.size() / (double)this.bmidmap.size()) +"\n" +
                "";
    }
}

