/*
 *  Copyright (c) 2003 Helsinki Institute for Information Technology
 *
 *  Permission is hereby granted, free of charge, to any person obtaining a copy
 *  of this software and associated documentation files (the "Software"), to
 *  deal in the Software without restriction, including without limitation the
 *  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 *  sell copies of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *
 *  The above copyright notice and this permission notice shall be included in
 *  all copies or substantial portions of the Software.
 *
 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 *  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 *  IN THE SOFTWARE.
 */

package vbsearch.query;

import com.hp.hpl.jena.rdf.model.Literal;
import java.util.*;
import vbsearch.*;
import vbsearch.util.*;

/**
 *  Description of the Class
 *
 *@author     Samppa
 *@created    September 17, 2003
 */
public class BasicQueryContext implements QueryContext {
    /**
     *  Description of the Field
     */
    private VBSIndex index;


    /**
     *  Constructor for the QueryContext object
     *
     *@param  index  Description of the Parameter
     */
    public BasicQueryContext(VBSIndex index) {
        this.index = index;
    }


    /**
     *  Gets the bookmarks attribute of the QueryContext object
     *
     *@param  cc  Description of the Parameter
     *@return     The bookmarks value
     */
    public Collection getBookmarks(CategoryConstraint cc) {
        final int EXPECTED_RESULT_SIZE = 256;
        Collection rv = null;
        Collection matchingCategories = getMatchingCategories(cc);
        // Bookmarks by categories
        if (matchingCategories != null) {
            if (rv == null) {
                rv = new LinkedHashSet(EXPECTED_RESULT_SIZE);
            }
            for (Iterator i = matchingCategories.iterator(); i.hasNext(); ) {
                //System.out.println("matchingCategory");
                for (Iterator j = ((Category) i.next()).getBookmarks(false); j.hasNext(); ) {
                    Bookmark bm = (Bookmark) j.next();
                    //System.out.print("bm ");
                    if (!rv.contains(bm)) {
                        rv.add(bm);
                    }
                }
            }
        }
        // Bookmarks by keyword (bookmarkKeyword)
        for (Iterator i = cc.getBookmarkKeyword(); i.hasNext(); ) {
            if (rv == null) {
                rv = new LinkedHashSet(EXPECTED_RESULT_SIZE);
            }
            for (Iterator j = this.index.getBookmarksByKeyword(((Literal) i.next()).getString()); j.hasNext(); ) {
                Bookmark bm = (Bookmark) j.next();
                if (!rv.contains(bm)) {
                    rv.add(bm);
                }
            }
        }
        // Known bookmarks <ogt:bookmark rdf:resource="...
        for (Iterator bmiter = cc.getBookmark(); bmiter.hasNext(); ) {
            if (rv == null) {
                rv = new LinkedHashSet(EXPECTED_RESULT_SIZE);
            }
            Bookmark bm = this.index.getBookmarkByURI((String) bmiter.next());
            if (bm != null && !rv.contains(bm)) {
                rv.add(bm);
            }
        }
        //System.out.println("\nBookmarks: " + rv.size());
        return rv;
    }


    /**
     *  Gets the intersection attribute of the QueryContext object
     *
     *@param  selector  Description of the Parameter
     *@return           The intersection value
     */
    public Collection getIntersection(FacetSelector selector) {
        Collection rv = null;
        for (Iterator cciter = selector.getConstraints(); cciter.hasNext(); ) {
            Collection bookmarks = getBookmarks((CategoryConstraint) cciter.next());
            if (bookmarks != null) {
                if (rv == null) {
                    rv = new LinkedHashSet(bookmarks);
                } else {
                    rv.retainAll(bookmarks);
                }
            }
        }
        return rv;
    }


    /**
     *  Gets the matchingCategories attribute of the FacetSelector object
     *
     *@param  cc  Description of the Parameter
     *@return     The matchingCategories value
     */
    public Collection getMatchingCategories(CategoryConstraint cc) {
        Collection categories = null;
        Collection infCategories = null;

        // Categories by ogt:catid --> categories
        for (Iterator i = cc.getCategory(); i.hasNext(); ) {
            Category category = this.index.getCategory(i.next().toString());
            if (categories == null) {
                categories = new LinkedList();
            }
            if (category != null) {
                categories.add(category);
            }
        }
        // Categories by keyword --> inferred categories
        for (Iterator i = cc.getProjectedFrom(); i.hasNext(); ) {
            if (infCategories == null) {
                infCategories = new LinkedList();
            }
            Iterator j = this.index.getCategoriesByProjection(i.next().toString(), categories);
            if (j != null) {
                while (j.hasNext()) {
                    infCategories.add(j.next());
                }
            }
        }
        // Switch categories and infCategories
        if (infCategories != null) {
            categories = infCategories;
            infCategories = null;
        }
        //System.out.println("Keywords? ");
        // Keyword --> inferred categories
        for (Iterator i = cc.getCategoryKeyword(); i.hasNext(); ) {
            if (infCategories == null) {
                infCategories = new LinkedList();
            }
            String kw = ((Literal) i.next()).getString();
            Iterator j = this.index.getCategoriesByKeyword(kw, categories);
            if (j != null) {
                while (j.hasNext()) {
                    infCategories.add(j.next());
                }
            }
        }
        if (infCategories != null) {
            categories = infCategories;
        }
        /*
         *  else if (categories == null) {
         *  categories = new LinkedList();
         *  for (Iterator i = this.index.getRoots(); i.hasNext(); ) {
         *  categories.add(i.next());
         *  }
         *  }
         */
        return categories;
    }
}

