/*
 *  Copyright (c) 2003 Helsinki Institute for Information Technology
 *
 *  Permission is hereby granted, free of charge, to any person obtaining a copy
 *  of this software and associated documentation files (the "Software"), to
 *  deal in the Software without restriction, including without limitation the
 *  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 *  sell copies of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *
 *  The above copyright notice and this permission notice shall be included in
 *  all copies or substantial portions of the Software.
 *
 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 *  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 *  IN THE SOFTWARE.
 */

package vbsearch.query;

import java.util.*;
import vbsearch.*;
import vbsearch.util.*;

/**
 *  Description of the Class
 *
 *@author     Samppa
 *@created    September 17, 2003
 */
public class CachedQueryContext extends BasicQueryContext {

    private final static boolean CACHE_BOOKMARKS = true;
    private final static boolean CACHE_INTERSECTIONS = true;
    private final static boolean CACHE_MATCHING_CATEGORIES = true;

    private HashMap bookmarks;

    private HashMap intersections;

    /**
     *  Description of the Field
     */
    private HashMap matchingCategories;


    /**
     *  Constructor for the QueryContext object
     *
     *@param  index  Description of the Parameter
     */
    public CachedQueryContext(VBSIndex index) {
        super(index);
        if (CACHE_BOOKMARKS) {
            this.bookmarks = new HashMap(256);
        }
        if (CACHE_INTERSECTIONS) {
            this.intersections = new HashMap(8);
        }
        if (CACHE_MATCHING_CATEGORIES) {
            this.matchingCategories = new HashMap(8);
        }
    }


    /**
     *  Gets the bookmarks attribute of the QueryContext object
     *
     *@param  cc  Description of the Parameter
     *@return     The bookmarks value
     */
    public Collection getBookmarks(CategoryConstraint cc) {
        if (CACHE_BOOKMARKS) {
            Collection rv = null;
            Object key = cc.isAnon() ? cc.getID() : cc.getURI();
            if (this.bookmarks.containsKey(key)) {
                rv = (Collection) this.bookmarks.get(key);
            } else {
                rv = super.getBookmarks(cc);
                this.bookmarks.put(key, rv);
            }
            return rv;
        } else {
            return super.getBookmarks(cc);
        }
    }


    /**
     *  Gets the intersection attribute of the QueryContext object
     *
     *@param  fs  Description of the Parameter
     *@return     The intersection value
     */
    public Collection getIntersection(FacetSelector fs) {
        if (CACHE_INTERSECTIONS) {
            Collection rv = null;
            Object key = fs.getIntesectionKey();
            // cacheable intersection?
            if (key != null) {
                if (this.intersections.containsKey(key)) {
                    rv = (Collection) this.intersections.get(key);
                } else {
                    rv = super.getIntersection(fs);
                    this.intersections.put(key, rv);
                }
            } else {
                rv = super.getIntersection(fs);
            }
            return rv;
        } else {
            return super.getIntersection(fs);
        }
    }


    /**
     *  Gets the matchingCategories attribute of the FacetSelector object
     *
     *@param  cc  Description of the Parameter
     *@return     The matchingCategories value
     */
    public Collection getMatchingCategories(CategoryConstraint cc) {
        if (CACHE_MATCHING_CATEGORIES) {
            Collection rv = null;
            Object key = cc.isAnon() ? cc.getID() : cc.getURI();
            if (this.matchingCategories.containsKey(key)) {
                rv = (Collection) this.matchingCategories.get(key);
            } else {
                rv = super.getMatchingCategories(cc);
                this.matchingCategories.put(key, rv);
            }
            return rv;
        } else {
            return super.getMatchingCategories(cc);
        }
    }
}

